#ifndef CQTREE_H
#define CQTREE_H

#include <vector>
#include <functional>
#include <fstream>

template<typename Vertex>
class Entity {
private:
	using Cord = std::pair< float, float >;
	float x_, y_;
	Vertex v_;
	int g_;

public:
	Entity() = default;
	Entity(Cord c, Vertex v, int g) {
		x_ = c.first;
		y_ = c.second;
		v_ = v;
		g_ = g;
	}
	double x() const {
		return x_;
	}
	double y() const {
		return y_;
	}
	Vertex v() const {
		return v_;
	}
	int g() const {
		return g_;
	}
	void setX(double x) {
		x_ = x;
	}

	friend bool operator == (Entity & lhs, Entity & rhs ) {
		return (lhs.x() == rhs.x() && lhs.y() == rhs.y() );
	}
};

template<typename Vertex>
class QTree{
private:
	using VVector = std::vector< Vertex >;
	using CVVector = std::vector<Vertex>;

	using VMatrix = std::vector< VVector >;
	using BNeighbors = std::vector< VVector* >;
	using Cord = std::pair< float, float >;


	const float x_, y_, whalf_, hhalf_;
	const unsigned MAX_N;
	float dLim;
	unsigned n_;
	bool stop, leaf;
	int nGraph_;

	std::queue< Entity<Vertex> > entities_;
	VMatrix vertices;

	std::unique_ptr< QTree<Vertex> > nw_, ne_, se_, sw_;

public:
	QTree(float x, float y, float width, float height, int bsize, int ngraph) : x_(x), y_(y), whalf_(width / 2), hhalf_(height / 2), MAX_N(bsize){
		n_ = 0;
		nGraph_ = ngraph;
		stop= false;
		leaf = true;
		dLim = 0.005;
		vertices.resize(ngraph, VVector(0));
	}

	void moveVertices(){
		float w4 = whalf_/2, h4 = hhalf_/2;

		nw_ = std::unique_ptr<QTree<Vertex>>(new QTree<Vertex>(x_ - w4, y_ - h4, whalf_, hhalf_, MAX_N, nGraph_));
		ne_ = std::unique_ptr<QTree<Vertex>>(new QTree<Vertex>(x_ + w4, y_ - h4, whalf_, hhalf_, MAX_N, nGraph_));
		se_ = std::unique_ptr<QTree<Vertex>>(new QTree<Vertex>(x_ + w4, y_ + h4, whalf_, hhalf_, MAX_N, nGraph_));
		sw_ = std::unique_ptr<QTree<Vertex>>(new QTree<Vertex>(x_ - w4, y_ + h4, whalf_, hhalf_, MAX_N, nGraph_));

		Entity<Vertex> ent;
		while(!entities_.empty()){
			ent = entities_.front();
			entities_.pop();
			if (ent.y() < y_) {
				if (ent.x() < x_) {
					nw_->insert(std::move(ent));
				} else {
					ne_->insert(std::move(ent));
				}
			} else {
				if (ent.x() < x_) {
					sw_->insert(std::move(ent));
				} else {
					se_->insert(std::move(ent));
				}
			}
		}
		n_ = 0;
	}

	void insert(Entity<Vertex>&& entity) {
		if( leaf ) {
			entities_.push( std::move( entity) );
			n_++;

			if( n_ == MAX_N  ) {
				if( whalf_/2 < dLim ){
					stop = true;
				} else {
					moveVertices();
					leaf = false;
				}
			}
		} else {
			auto ent = std::move(entity);
			if (ent.y() < y_) {
				if (ent.x() < x_) {
					nw_->insert(std::move(ent));
				} else {
					ne_->insert(std::move(ent));
				}
			} else {
				if (ent.x() < x_) {
					sw_->insert(std::move(ent));
				} else {
					se_->insert(std::move(ent));
				}
			}
		}
	}

	void transfer( int g ) {
		if( vertices[g].size() == 0 ){
			Entity<Vertex> ent;
			while( !entities_.empty() ) {
				ent = entities_.front();
				entities_.pop();
				vertices[g].push_back( ent.v() );
			}
		}
	}

	VVector* lookup( Cord& p, Cord& c, double& r, int g ) {
		float x = p.first;
		float y = p.second;

		if ( leaf ) {
			r = radius()/2;
			c = Cord((float)x_, (float)y_);
			transfer(g);
			return &vertices[g];
		} else {
			if (y < y_) {
				if (x < x_) {
					return nw_->lookup(p, c, r, g);
				} else {
					return ne_->lookup(p, c, r, g);
				}
			} else {
				if (x < x_) {
					return sw_->lookup(p, c, r, g);
				} else {
					return se_->lookup(p, c, r, g);
				}
			}
		}
	}

	Cord getCenter( Cord p, double& r ) {
		int g; Cord c;
		lookup( p, c, r, 0 );
		return c;
	}

	void nlookup(Cord& p, Cord& c, double& r, BNeighbors& result, int g) {
		if ( !leaf ) {
			nw_->nlookup(p, c, r, result, g);
			ne_->nlookup(p, c, r, result, g);
			se_->nlookup(p, c, r, result, g);
			sw_->nlookup(p, c, r, result, g);
		} else {
			double rcurr = radius()/2;

			bool xBorder = (std::abs(c.first-x_) <= rcurr+r);
			bool yBorder = (std::abs(c.second-y_) <= rcurr+r);

			if( xBorder && yBorder ) {
				transfer(g);
				if(vertices[g].size() >0 ) {
					result.push_back( &vertices[g] );
				}
			}
		}
	}

	std::pair< VVector*, BNeighbors > neighbors( Cord p, int o )
	{
		BNeighbors neighborBuckets;
		Cord c;
		double r,t;

		VVector* v1 = lookup( p, c, r, o );

		nlookup( p, c, r, neighborBuckets, (o==0) );

		return std::make_pair( v1, neighborBuckets );
	}

	BNeighbors neighborsOf( Cord p )
	{
		BNeighbors neighborBuckets;
		double r,t;
		Cord c = getCenter( p, r );
		nlookup( p, c, r, neighborBuckets, 0 );

		return neighborBuckets;
	}

	std::pair< VVector*, BNeighbors > neighborsOf(  )
	{
		BNeighbors neighborBuckets;
		Cord c, p(x_, y_);
		double r,t;

		VVector* v1 = &vertices[1];
		VVector* v2 = &vertices[2];

		neighborBuckets.push_back( v2 );

		nlookup( p, c, r, neighborBuckets, 0);

		return std::make_pair( v1, neighborBuckets );
	}

	void traverse(std::function<void(QTree<Vertex>&, unsigned, int*)> fn, unsigned depth, int* match) {
		fn(*this, depth, match);
		if ( !leaf ) {
			nw_->traverse(fn, depth + 1, match);
			ne_->traverse(fn, depth + 1, match);
			se_->traverse(fn, depth + 1, match);
			sw_->traverse(fn, depth + 1, match);
		}
	}

	VVector* getBucket() {
		transfer(0);
		return &vertices[0];
	}

	Cord getCenter() {
		return Cord( x_, y_ );
	}

	bool isLeaf() {
		return leaf;
	}

	void print() {
		std::cout << vertices[0].size() << " " << vertices[1].size() << std::endl;
	}

	float DistEuclidien(Cord s, Cord d)
	{
		return (float) sqrt( (s.first-d.first)*(s.first-d.first) + (s.second-d.second)*(s.second-d.second) );
	}

	float radius() const {
		return 2*whalf_;
	}

	int getBucketSize() {
		int bucketSize=0;
		for( int i=0 ; i<vertices.size() ; i++ ) {
			auto v = vertices[i];
			bucketSize += v.size();
		}

		return bucketSize;
	}

	void printQtree( std::ostream& o ) {

		if( leaf ) {
			o << x_ << " " << y_ << " " << whalf_ << " " << getBucketSize() << std::endl;
		} else {
			nw_->printQtree( o );
			ne_->printQtree( o );
			sw_->printQtree( o );
			se_->printQtree( o );
		}
	}
};


#endif