#include <iostream>
#include <fstream>
#include <random>
#include <set>
#include <algorithm>
#include <iomanip>
#include <atomic>
#include <chrono>
#include <unordered_map>
#include <unistd.h>

#include "graph.hpp"
#include "gtruth.hpp"
#include "anchors.hpp"
#include "order.hpp"
#include "position.hpp"
#include "similarity.hpp"
#include "merger.hpp"
#include "timing.hpp"

#include "cqtree.hpp"

#define DEV_LOG

using Vertex = int;
using TWPair = std::pair<float, int>;
using Edge = std::pair< Vertex, float >;
using EType = std::pair< Vertex, int >;
using IPair = std::pair<int, int>;
using VectIPair = std::vector<IPair>;
using Cord = std::pair< float, float >;
using sGraph = Graph<Vertex, Edge, EType>;
using sAnchor = Anchors<Vertex, Edge, EType>;
using Placer = GlobalPlacer< Vertex, Edge, EType >;

using Tree = QTree<Vertex>;
using VVector = std::vector< Vertex >;
using BNeighbors = std::vector< VVector* >;
using Task = std::pair< VVector*, BNeighbors >;
using TaskP = std::pair< VVector*, VVector*>;
using LTask = std::vector< Task >;
using LPTask = std::vector< TaskP >;
using sMap = std::vector< std::pair<Vertex, float> >;

using namespace std::chrono;

// I/O functions
std::string readGraph(std::fstream& confIn, sGraph& g);
void readConfFile( std::string f, sGraph& A, sGraph& B, GTruth<Vertex>& gt );

// QTree
void insertVertices( sGraph& A, Placer& p, Tree& t, int o, int skip );
//void compTasks( Tree& t, LTask& l );
void compTasks( Tree& t1, Tree& t2, LTask& l );

float getInSec(high_resolution_clock::time_point t1, high_resolution_clock::time_point t2);

float test(sGraph &A, sGraph &B, sAnchor &aA, sAnchor &aB, Placer &aP, Placer &bP, GTruth<Vertex> &gt, int dLim);
float PSM(sGraph &A, sGraph &B, sAnchor &aA, sAnchor &aB, Placer &aP, Placer &bP);


int main(int argc, const char * argv[])
{
	Timing times;
	int taskType = atoi( argv[7] );

	high_resolution_clock::time_point t1,t2;

	std::string confFile = argv[1];
	int k = atoi( argv[3] );
	const int bucketSize = atoi( argv[2] );
	const float width=2, height=2;
	int nChanges = 2<<22, nIters=0;

	/* Read graphs */
	Graph<Vertex, Edge, EType> A;
	Graph<Vertex, Edge, EType> B;
	GTruth<Vertex> gt;

	readConfFile(confFile, A, B, gt);

	/* Prepare Anchors */

#if defined(C_EXEC) || defined(M_EXEC)
	t1 = high_resolution_clock::now();
#endif
	Anchors<Vertex, Edge, EType> aA( k, A, gt );
	Anchors<Vertex, Edge, EType> aB( aA.getAnchors(), gt );

#if defined(C_EXEC) || defined(M_EXEC)
	t2 = high_resolution_clock::now();
	times.insert( "initial anchor", getInSec(t1, t2) );
#endif


	/* distance matrix */
#if defined(C_EXEC) || defined(M_EXEC)
	t1 = high_resolution_clock::now();
#endif
	A.setDistanceMatrix( aA.getAnchors() );
	B.setDistanceMatrix( aB.getAnchors() );

#if defined(C_EXEC) || defined(M_EXEC)
	t2 = high_resolution_clock::now();
	times.insert( "initial bfs", getInSec(t1, t2) );
#endif

	int simLimit = atoi( argv[4] );
	Merger<Vertex, Edge, EType > merge( &A, &B, &aA, &aB, simLimit );
	sMap m( A.nVert()+1, Edge(-1,-1) );
	sMap mR( B.nVert()+1, Edge(-1,-1) );
	sMap mP;

	Organizer<Vertex, Edge, EType> org;
	VectIPair order;

	auto rng = std::default_random_engine {};

	while( nChanges > B.nVert()*0.02 && nIters<10 ) {
		org.init( A, B, aA, aB );
#if defined(C_EXEC) || defined(M_EXEC)
		t1 = high_resolution_clock::now();
#endif
		order = org.organizeAnchors();

#if defined(C_EXEC) || defined(M_EXEC)
		t2 = high_resolution_clock::now();
		times.insert( "order", getInSec(t1, t2) );
#endif

		/* Vantage anchor computation */
		int dLim = atoi(argv[5]);
		Placer gPlacerA, gPlacerB;

		gPlacerA = Placer( A, aA, order );
		gPlacerB = Placer( B, aB, order );

#if defined(C_EXEC) || defined(M_EXEC)
		t1 = high_resolution_clock::now();
#endif
		Tree qtA( width/2, height/2, width, height, bucketSize, 1 );
		Tree qtB( width/2, height/2, width, height, bucketSize, 1 );

		int skipDeg = (nIters < 1) ? 2 : -1;
		{insertVertices( A, gPlacerA, qtA, 0, skipDeg );}
		{insertVertices( B, gPlacerB, qtB, 1, skipDeg );}

#if defined(C_EXEC) || defined(M_EXEC)
		t2 = high_resolution_clock::now();
		times.insert( "qtree insertion", getInSec(t1, t2) );
#endif


		LTask tasks;
		compTasks( qtA, qtB, tasks );

		mP = m;
		for( int i=0 ; i<aA.getAnchorSize() ; i++ ){
			m[ aA.getAnchor(i) ] = Edge( aB.getAnchor(i), 1.01 );
			mR[ aB.getAnchor(i) ] = Edge( aA.getAnchor(i), 1.01 );
		}

#ifdef C_EXEC
		t1 = high_resolution_clock::now();
#endif

		merge.map( tasks, m, mR );
		// tasks.clear();

#ifdef C_EXEC
		t2 = high_resolution_clock::now();
		std::clog<< "map graph : " << getInSec(t1, t2) << std::endl;
#endif

		nChanges = 0;

#ifdef C_EXEC
		t1 = high_resolution_clock::now();
#endif
		for( int i=0 ; i<m.size() ; i++ ) {
			nChanges += (m[i].first != mP[i].first);
		}

#ifdef C_EXEC
		t2 = high_resolution_clock::now();
		std::clog<< "count changes : " << getInSec(t1, t2) << std::endl;
#endif

#ifdef M_EXEC
		merge.printExecTimes( times );
#endif
		++nIters;
	}

	std::cout << "Recall : " << gt.getRecall(m) << std::endl;
	times.printAll();
	return 0;
}

float DistEuclidien(Cord s, Cord d) {
	return (float)std::sqrt( (s.first-d.first)*(s.first-d.first) + (s.second-d.second)*(s.second-d.second) );
}

void insertVertices( sGraph& A, Placer& p, Tree& t, int o, int skip ){
	for( int i=0 ; i<A.nVert() ; i++ ) {
		if( skip && A.getDeg(i)<skip ) continue;
		Cord pos = p.trianglePosOnSamples( i );
		t.insert( Entity<Vertex>(pos, i, 0) );
	}
}

void compTasks( Tree& t1, Tree& t2, LTask& l ) {
	int miss =0;
	t2.traverse([&](Tree &tc, unsigned depth, int *ma) {
		if( tc.isLeaf() ){
			l.push_back( Task( tc.getBucket(), t1.neighborsOf( tc.getCenter() ) ) );
		}
	}, 0, &miss);
}

std::string readGraph(std::fstream& confIn, Graph<Vertex, Edge, EType>& g) {
	std::string GFile, TFile, MFile, AFile;
	confIn >> GFile >> TFile >> MFile >> AFile;

	// Read Graph
	std::fstream aG(GFile);
	int t; float w;
	Vertex u, v;

	while( aG >> u >> v >> w >> t ) {
		g.addEdge( u, Edge(v, w) );
		g.setEType( u, EType(v, t) );
	}
	aG.close();

	// Read Vertex types
	std::fstream aT(TFile);
	while( aT >> v >> t ) {
		g.setVType( v, t );
	}
	aT.close();

	// Read Vertex attributes
	std::string key, attr;
	std::fstream aA(AFile);
	while( aA >> key >> v >> attr ) {
		g.setAttribute( v, attr );
	}
	aA.close();

	g.setSWeights();
	g.setSTypes();
	return MFile;
}

void readConfFile( std::string f, Graph<Vertex, Edge, EType>& A, Graph<Vertex, Edge, EType>& B, GTruth<Vertex>& gt ) {
	std::fstream confIn(f);

	readGraph( confIn, A );
	std::string tfile = readGraph( confIn, B );
	gt.setGTruth( tfile, A.nVert() );
}

float getInSec(high_resolution_clock::time_point t1, high_resolution_clock::time_point t2) {
	return (float)duration_cast<microseconds>(t2-t1).count()/1000000;
}
