#ifndef MERGER_H
#define MERGER_H

#include <vector>
#include <iostream>
#include <queue>
#include <algorithm>
#include <set>
#include <mutex>

#include "graph.hpp"
#include "similarity.hpp"
#include "timing.hpp"

#include <cilk/cilk.h>

template<typename Vertex, typename Edge, typename EType>
class Merger{
private:
	class CompareF {
		public:
			bool operator()(const Edge& u, const Edge& v) const {
				return u.second < v.second;
			}
	};

	struct Order {
	    bool operator()( Edge const& a, Edge const& b ) const {
	        return a.second > b.second;
	    }
	};

	using sGraph = Graph<Vertex, Edge, EType>;
	using sAnchors = Anchors<Vertex, Edge, EType>;
	using VVector = std::vector< Vertex >;
	using BNeighbors = std::vector< VVector* >;
	using Task = std::pair< VVector*, BNeighbors >;
	using LTask = std::vector< Task >;
	using TaskP = std::pair< VVector*, VVector* >;
	using LPTask = std::vector< TaskP >;
	using PSim = PairSim< Vertex, Edge, EType >;
	using Queue = std::vector< Edge >;
	using PQueue = std::priority_queue<Edge, std::vector<Edge>, Order >;
	using sMatrix = std::vector< Queue >;
	using sMap = std::vector< std::pair<Vertex, float> >;

	sGraph* A, *B;
	sAnchors* aA, *aB;
	PSim ps;
	sMatrix sM;
	int simLimit, aC, aC0;
	std::mutex mtx;

#ifdef M_EXEC
	float tPrep, tMap, tSim;
	high_resolution_clock::time_point t1,t2;

	float getInSec() {
		return (float)duration_cast<microseconds>(t2-t1).count()/1000000;
	}
#endif

	void compForVNeighbor( Vertex v, VVector* n, sMap& m, sMap& mR ) {
		auto& temp = sM[v];//Queue( simLimit, Edge(-1,-1) );
		if( temp.size()==0 ) {
			temp.resize(simLimit, Edge(-1,-1));
		}

		for( int i=0 ; i<n->size() ; i++ ) {
			auto u = n->at(i);
			float sim = ps.sigma(u, v, m, mR);

			auto top = temp[0];
			if( findAndUpdate(temp, u, sim)==false &&
					sim > top.second ) {
				int j=0;
				while( temp[j+1].second<sim && j<simLimit-1 ) {
					temp[j].first = temp[j+1].first;
					temp[j].second = temp[j+1].second;
					++j;
				}
				temp[j].first = u;
				temp[j].second = sim;
			}

		}

	}

	void compForVNeighbor3( Vertex v, VVector* n, sMap& m, sMap& mR ) {
		auto temp = Queue( simLimit, Edge(-1,-1) );
		for( int i=0 ; i<n->size() ; i++ ) {
			auto u = n->at(i);
			float sim = ps.sigma(u, v, m, mR);

			auto top = temp[0];
			if( findAndUpdate(temp, u, sim)==false &&
					sim > top.second ) {
				int j=0;
				while( temp[j+1].second<sim && j<simLimit-1 ) {
					temp[j].first = temp[j+1].first;
					temp[j].second = temp[j+1].second;
					++j;
				}
				temp[j].first = u;
				temp[j].second = sim;
			}
		}
		auto& sV = sM[v];
		auto ts = (sV.size()>0) ? sV[0].second : -1;
		if( temp.back().second>ts ) {
			if( sV.size()==0 ) {
				sV.resize(simLimit, Edge(-1,-1));
			}
			int i,j,k;
			i=j=k=simLimit-1;

			Edge max(-1,-1);
			Queue topK( simLimit, Edge(-1,-1) );

			while( k>=0 ) {
				if( temp[i].second > sV[j].second ) {
					topK[k] = temp[i];
					i--;
				} else {
					topK[k] = sV[j];
					j--;
				}
				k--;
			}
			for( int i=0 ; i<simLimit ; i++ ) {
				sV[i] = topK[i];
			}
		}
	}

	bool findAndUpdate( Queue& q, Vertex v, float sim ) {
		bool updated = false;
		int i=0;
		if( q[0].first>=0 ) {
			while( q[i].first!=v && i<q.size() ){
				++i;
			}
			if( i<q.size() ) {
				q[i].second = std::max( q[i].second, sim );
				updated = true;
			}
		}
		return updated;
	}
public:
	Merger( ) { }

	Merger( sGraph* A, sGraph* B, sAnchors* aA, sAnchors* aB, int simLimit ) {
		this->A = A;
		this->B = B;
		this->aA = aA;
		this->aB = aB;
		this->simLimit = simLimit;
		sM.resize( B->nVert(), Queue(0) );
		aC=aA->getAnchorSize()+1;
		aC0 = aA->getAnchorSize();

		tMap = tSim = tPrep = 0;
	}

	void compSim( LTask& task, sMap& m, sMap& mR ) {

		cilk_for( int i=0 ; i<task.size() ; i++ ) {
		// for each task
			auto b = task[i];
			// for every vertex in the bucket of graph 2
			for( int j=0 ; j<b.first->size() ; j++ ) {
				auto v = b.first->at(j);
				// in every neighbour
				for( int k=0 ; k<b.second.size() ; k++ ) {
					auto n = b.second[k];
					// compare it with the vertices from graph 1
					compForVNeighbor( v, n, m, mR );
				}
			}
		}
	}

	void compSim( LPTask& task, sMap& m, sMap& mR ) {
#ifdef M_OMP
		#pragma omp parallel for schedule(dynamic,1)
		for( int i=0 ; i<task.size() ; i++ ) {
#elif M_CILK
		cilk_for( int i=0 ; i<task.size() ; i++ ) {
#else
		for( int i=0 ; i<task.size() ; i++ ) {
#endif
			for( int j=0 ; j<task[i].first->size() ; j++ ) {
				auto v = task[i].first->at(j);
				auto n = task[i].second;
				compForVNeighbor3( v, n, m, mR );
			}
		}
	}

	void mapVertices( sMap& m, sMap& mR ) {
		int n = B->nVert();
		for( int v=0 ; v<n ; v++ ) {
			if( sM[v].size()>0 ) {
				place( v, m, mR );
			}
		}
	}

	void place( Vertex v2, sMap& m, sMap& mR ) {
		if( mR[v2].second > 0 )
			return;

		auto top = sM[v2].back();
		Vertex v = top.first;
		float sim = top.second;
		sM[v2].pop_back();

		if( m[v].second < sim ) {
			Vertex p = m[v].first;

			m[v] = Edge(v2, sim);
			mR[v2] = top;
			if( p >= 0 ) {
				mR[p] = Edge(-1, -1);
				if( sM[p].size()>0 )
					place( p, m, mR );
			}
		}
	}

	void map( LTask& task, sMap& m, sMap& mR ) {
		ps.init( *A, *B, *aA, *aB );

#ifdef M_EXEC
		t1 = high_resolution_clock::now();
#endif
		compSim( task, m, mR );
#ifdef M_EXEC
		t2 = high_resolution_clock::now();
		tSim = getInSec();
#endif

#ifdef M_EXEC
		t1 = high_resolution_clock::now();
#endif

		mapVertices( m, mR );

#ifdef M_EXEC
		t2 = high_resolution_clock::now();
		tMap = getInSec();
#endif

#ifdef M_EXEC
		t1 = high_resolution_clock::now();
#endif
		newAnchors(m, mR);

#ifdef M_EXEC
		t2 = high_resolution_clock::now();
		tPrep = getInSec();
#endif
	}

	void map( LPTask& task, sMap& m, sMap& mR ) {
		ps.init( *A, *B, *aA, *aB );

#ifdef M_EXEC
		t1 = high_resolution_clock::now();
#endif
		compSim( task, m, mR );
#ifdef M_EXEC
		t2 = high_resolution_clock::now();
		tSim = getInSec();
#endif

#ifdef M_EXEC
		t1 = high_resolution_clock::now();
#endif

		mapVertices( m, mR );

#ifdef M_EXEC
		t2 = high_resolution_clock::now();
		tMap = getInSec();
#endif

#ifdef M_EXEC
		t1 = high_resolution_clock::now();
#endif
		newAnchors(m, mR);

#ifdef M_EXEC
		t2 = high_resolution_clock::now();
		tPrep = getInSec();
#endif
	}

	void newAnchors( sMap& m, sMap& mR ) {
		if( aC > aA->getAnchorSize() ) {
			PQueue q;
			for( int i=0 ; i<m.size() ; i++ ) {
				if( m[i].second <= 1 && m[i].first>=0 ) {
					q.push( m[i] );
					if( q.size() >= aC ) {
						q.pop();
					}
				}
			}
			int aSize=aA->getAnchorSize();

			while( !q.empty() ) {
				auto t = q.top();
				q.pop();

				Vertex v2 = t.first;
				Vertex v1 = mR[v2].first;

				mR[ v2 ].second = 1.01;
				m[ v1 ].second = 1.01;

				aA->addAnchor( v1 );
				A->addDistanceVector( v1 );

				aB->addAnchor( v2 );
				B->addDistanceVector( v2 );
			}
#ifdef M_OMP
			#pragma omp parallel for schedule(dynamic,2)
#endif
			for( int i=aSize ; i<aA->getAnchorSize() ; i++ ) {
					A->BFS( aA->getAnchor(i), i );
					B->BFS( aB->getAnchor(i), i );
			}

			aC = ( aC<1000 ) ? 2*aC : aC0;
		} else {
			while( aA->getAnchorSize() > aC0 ) {
				Vertex v1 = aA->getAnchor( aA->getAnchorSize()-1 );
				Vertex v2 = aB->getAnchor( aA->getAnchorSize()-1 );

				float sim = ps.sigmaBase( v1, v2 );

				m[v1].second = sim;
				mR[v2].second = sim;

				aA->eraseLastAnchor();
				aB->eraseLastAnchor();

				A->eraseLastAnchor();
				B->eraseLastAnchor();
			}
			aC++;
		}
	}
#ifdef M_EXEC
	void printExecTimes() {
		std::clog << "sim time : " << tSim << std::endl;
		std::clog << "map time : " << tMap << std::endl;
		std::clog << "preperation time : " << tPrep << std::endl;
	}

	void printExecTimes( Timing& t) {
		t.insert( "sim time", (float)tSim );
		t.insert( "map time", (float)tMap );
		t.insert( "preperation time", (float)tPrep );
	}
#endif
};


#endif