#ifndef ORDER_H
#define ORDER_H

#include <vector>
#include <iostream>
#include <queue>
#include <algorithm>
#include <cmath>
#include <unordered_set>
#include <random>

#include "graph.hpp"
#include "anchors.hpp"

template<typename Vertex, typename Edge, typename EType>
class Organizer{
private:
	struct Greater
	{
	    bool operator()(EType const& a, EType const& b) const
	    {
	        return a.second < b.second;
	    }
	};

	using GreaterQ = std::priority_queue<EType, std::vector<EType>, Greater >;
	using VertexM = std::vector< std::vector<Vertex> >;
	using sGraph = Graph<Vertex, Edge, EType>;
	using sAnchors = Anchors<Vertex, Edge, EType>;
	using IPair = std::pair<int, int>;
	using VectIPair = std::vector<IPair>;

	sGraph A,B;
	sAnchors aA, aB;

	std::vector< Vertex > centralAnchors;
	std::vector< Vertex > vantageAnchors;
	VectIPair orderedVAnchors;
	std::unordered_set< Vertex > ignore;

	std::random_device rd;  //Will be used to obtain a seed for the random number engine
	std::mt19937 gen;//(rd()); //Standard mersenne_twister_engine seeded with rd()
	std::uniform_int_distribution<> dis;

	void topDeg( GreaterQ& q, int cSize ) {
		int n = aA.getAnchors().size();

		for( int i=0 ; i<n ; i++ ) {
			Vertex a = aA.getAnchor( i );
			int deg = A.getDeg( a );

			if( q.size() < cSize ) {
				q.push( EType(i, deg) );
			}
			else if( q.top().second < deg ) {
				q.pop();
				q.push( EType(i, deg) );
			} else {
				continue;
			}
		}
	}

	void chooseCenters( GreaterQ& q, int cSize ){
		while( !q.empty() && cSize>0 ) {
			bool central = true;
			int o = q.top().first;
			Vertex v = aA.getAnchor( o );

			/* eliminate close ones */
			int i=0;
			while( central && i<centralAnchors.size() )	{
				auto c = centralAnchors[i];
				central = ( A.getDistance(c,v)>1 );
				++i;
			}
			if( central ) {
				centralAnchors.push_back( o );
			} else {
				ignore.insert( o );
			}
			q.pop();
			--cSize;
		}
	}

	int findCloseCenters( VertexM& cluster, int& cLim ) {
		int nAnch=0;

		std::unordered_set<Vertex> caSet;
		for( int i=0 ; i<centralAnchors.size() ; i++ ) {
			auto c = centralAnchors[i];
			caSet.insert( c );
		}

		for( int i=0 ; i<aA.getAnchors().size() ; i++ ) {
			if( caSet.find(i) != caSet.end() ) {
				continue;
			}

			Vertex a = aA.getAnchor( i );
			int min = 2<<22, cInd=-1;

			for( int j=0 ; j<centralAnchors.size() ; j++ ) {
				auto c = centralAnchors[j];
				int dist = A.getDistance(c, a);

				if( dist < min ) {
					min = dist;
					cInd = c;
				}
			}

			if( cInd >= 0 ) {
				cluster[cInd].push_back( i );
				++nAnch;
			}
		}

		cLim = 2<<22;
		for( int i=0 ; i<cluster.size() ; i++ ){
			auto c = cluster[i];
			if ( c.size()>0 && c.size()<cLim )
				cLim = c.size();
		}

		return nAnch;
	}

	bool canAdd( std::vector<int>& counts, VertexM& cluster, int l ) {
		bool check = false;
		for( int i=0 ; i<counts.size() ; i++ ) {
			if( counts[i]<l && cluster[i].size()>0 ){
				check=true;
			}
		}

		return check;
	}

	void swap( std::vector< Vertex >& l, int i, int j ){
		int temp = l[i];
		l[i] = l[j];
		l[j] = temp;
	}

	bool sameInTwo( int i, int j ){
		int d1 = A.getDistance(i, aA.getAnchor(j));
		int d2 = B.getDistance(i, aB.getAnchor(j));

		return d1==d2;
	}
public:
	Organizer() {
		gen = std::mt19937(rd()); //Standard mersenne_twister_engine seeded with rd()
		dis = std::uniform_int_distribution<>(0, A.nVert());
	}

	Organizer( sGraph A, sGraph B, sAnchors aA, sAnchors aB ) {
		this->A = A;
		this->B = B;
		this->aA = aA;
		this->aB = aB;
		gen = std::mt19937(rd()); //Standard mersenne_twister_engine seeded with rd()
		dis = std::uniform_int_distribution<>(0, A.nVert());
	}

	void init( sGraph& A, sGraph& B, sAnchors& aA, sAnchors& aB ) {
		this->A = A;
		this->B = B;
		this->aA = aA;
		this->aB = aB;

		centralAnchors.clear();
		vantageAnchors.clear();
		orderedVAnchors.clear();
		ignore.clear();
	}

	void findCentralAnchors() {
		int cSize = (int) std::ceil(std::log( aA.getAnchors().size() ) / std::log(2));

		GreaterQ q;
		topDeg( q, cSize );
		chooseCenters( q, cSize );
	}

	void findVantageAnchors() {
		int cSize = centralAnchors.size();
		VertexM cluster( cSize, std::vector<Vertex>(0) );

		int cLim = 0, c=0;
		int nAnch = findCloseCenters( cluster, cLim );

		std::vector< int > counts( centralAnchors.size(), 0 );
		int v1, v2=-1;

		while( canAdd(counts, cluster, cLim) ) {
			if( cluster[c].size()>0 && counts[c]<cLim )
			{
				if( vantageAnchors.size()%2 == 0 ) {
					int ind = dis(gen) % cluster[c].size();
					v1 = cluster[c][ind];
					vantageAnchors.push_back(v1);
					cluster[c].erase( cluster[c].begin()+ind );
					counts[c] += 1;
				} else {
					int dist = 0;
					Vertex u = aA.getAnchor( v1 );
					for( int i=0 ; i<cluster[c].size() ; i++ ) {
						int v = cluster[c][i];
						int d = A.getDistance(v, u);
						if( d>dist && sameInTwo(v1,v) ) {
							dist = d;
							v2 = i;
						}
					}
					if( v2 < 0 ){
						vantageAnchors.pop_back();
					} else {
						vantageAnchors.push_back( cluster[c][v2] );
						cluster[c].erase( cluster[c].begin()+v2 );
						counts[c] += 1;
						v2=-1;
					}
				}
			}
			c = (c+1) % centralAnchors.size();
		}

		if( vantageAnchors.size() % 2 == 1 ) {
			vantageAnchors.pop_back();
		}
	}

	void orderVantageAnchors() {
		int vSize = vantageAnchors.size();
		for( int i=2 ; i<vSize-1 ; i+=2 ) {
			Vertex v = aA.getAnchor( vantageAnchors[i-2] );
			int min=2<<22, minInd=-1;

			for( int j=i ; j<vSize-1 ; j+=2 ) {
				int dist = A.getDistance( vantageAnchors[j], v );

				if( dist < min ) {
					min = dist;
					minInd = j;
				}
			}

			if( minInd !=i ) {
				swap( vantageAnchors, i, minInd );
				swap( vantageAnchors, i+1, minInd+1 );
			}
		}
	}

	VectIPair organizeAnchors() {
		findCentralAnchors();
		findVantageAnchors();
		orderVantageAnchors();

		for( int i=0 ; i<vantageAnchors.size()-1 ; i+=2 ) {
			int v1 = vantageAnchors[i];
			int v2 = vantageAnchors[i+1];
			orderedVAnchors.push_back( IPair(v1, v2) );
		}

		return orderedVAnchors;
	}
};


#endif