#ifndef POSITION_H
#define POSITION_H

#include <vector>
#include <iostream>
#include <queue>
#include <algorithm>
#include <cmath>
#include <unordered_set>
#include <random>

#include <boost/geometry.hpp>
#include <boost/geometry/geometries/point_xy.hpp>
#include <boost/geometry/geometries/polygon.hpp>
#include <boost/geometry/algorithms/centroid.hpp>

#include "graph.hpp"
#include "anchors.hpp"

#define PI 3.14159265

template<typename Vertex, typename Edge, typename EType>
class GlobalPlacer{
private:
	using sGraph = Graph<Vertex, Edge, EType>;
	using sAnchors = Anchors<Vertex, Edge, EType>;
	using IPair = std::pair<int, int>;
	using VectIPair = std::vector<IPair>;
	using Cord = std::pair< float, float >;

	typedef boost::geometry::model::d2::point_xy<float> point_type;
	typedef boost::geometry::model::polygon<point_type> polygon_type;

	sGraph A;
	sAnchors aA;

	VectIPair orderedVAnchors;
	std::vector< VectIPair > sampledOrder;

	std::mt19937_64 rng;
    std::uniform_real_distribution<double> unif;

	void setSampledOrder( int n ) {
		sampledOrder.resize( n );

		for( int i=0 ; i<orderedVAnchors.size() ; i++ ) {
			int j = i%n;
			sampledOrder[j].push_back( orderedVAnchors[i] );
		}
	}
public:
	GlobalPlacer() {}

	GlobalPlacer( sGraph A, sAnchors aA, VectIPair o ) {
		this->A = A;
		this->aA = aA;
		orderedVAnchors = o;

#ifdef EXPERIMENTAL
		setSampledOrder( orderedVAnchors.size()/8 );
#else
		setSampledOrder( 1 );
#endif
		uint64_t timeSeed = std::chrono::high_resolution_clock::now().time_since_epoch().count();
		std::seed_seq ss{uint32_t(timeSeed & 0xffffffff), uint32_t(timeSeed>>32)};
		rng.seed(ss);
		unif = std::uniform_real_distribution<double>(0,0.003);
	}

	Cord trianglePos( Vertex v ) {
		float RotAlpha = PI/orderedVAnchors.size();
		std::vector<Cord> sides;

		for( int i=0 ; i<orderedVAnchors.size() ; i++ ) {
			Cord st = Cord(1,0);
			Cord dt = Cord(-1,0);
			int a1 = orderedVAnchors[i].first;
			int a2 = orderedVAnchors[i].second;

			float fromS = A.getDistance( a1, v );
			float fromD = A.getDistance( a2, v );
			float anchD = A.getDistance( a1, aA.getAnchor(a2) );

			float cosalpha = -2;
			if( fromS==0 ) {
				cosalpha = 1;
			} else if( fromD==0 ) {
				cosalpha = -1;
			} else{
				cosalpha = (fromS*fromS+anchD*anchD-fromD*fromD)/(2*fromS*anchD);
			}

			float alpha = std::acos( cosalpha );

			float ratio = (DistEuclidien(st,dt)/anchD)*fromS;
			float x = st.first-std::cos(alpha)*ratio;
			float y = std::sin(alpha)*ratio;

			float nx = x*std::cos(i*RotAlpha) - y*std::sin(i*RotAlpha);
			float ny = x*std::sin(i*RotAlpha) + y*std::cos(i*RotAlpha);

			sides.push_back(Cord(nx,ny));
		}

		Cord pos = PolyCentroid(sides);

		return pos;
	}

	Cord trianglePos( Vertex v, int s ) {
		float RotAlpha = PI/sampledOrder[s].size();
		std::vector<Cord> sides;

		for( int i=0 ; i<sampledOrder[s].size() ; i++ ) {
			Cord st = Cord(1,0);
			Cord dt = Cord(-1,0);
			int a1 = sampledOrder[s][i].first;
			int a2 = sampledOrder[s][i].second;

			float fromS = A.getDistance( a1, v );
			float fromD = A.getDistance( a2, v );
			float anchD = A.getDistance( a1, aA.getAnchor(a2) );

			float cosalpha = -2;
			if( fromS==0 ) {
				cosalpha = 1;
			} else if( fromD==0 ) {
				cosalpha = -1;
			} else{
				cosalpha = (fromS*fromS+anchD*anchD-fromD*fromD)/(2*fromS*anchD);
			}

			float alpha = std::acos( cosalpha );

			float ratio = (DistEuclidien(st,dt)/anchD)*fromS;
			float x = st.first-std::cos(alpha)*ratio + 1./(A.getDeg(v));
			float y = std::sin(alpha)*ratio+ 1./(A.getDeg(v));

			float nx = x*std::cos(i*RotAlpha) - y*std::sin(i*RotAlpha);
			float ny = x*std::sin(i*RotAlpha) + y*std::cos(i*RotAlpha);

			sides.push_back(Cord(nx,ny));
		}

		Cord pos = PolyCentroid(sides);

		return pos;
	}

	Cord trianglePosOnSamples( Vertex v ) {
		std::vector<Cord> centroids;
		for( int i=0 ; i<sampledOrder.size() ; i++ ) {
			centroids.push_back( trianglePos(v, i) );
		}
		float x=0, y=0;
		for( int i=0 ; i<centroids.size() ; i++ ) {
			auto c = centroids[i];
			x += c.first;
			y += c.second;
		}
		x = x/sampledOrder.size();
		y = y/sampledOrder.size();

		return Cord(x+unif(rng),y+unif(rng));
	}

	Cord trianglePosWErr( Vertex v, int s, std::uniform_int_distribution<>& dis, std::mt19937& gen ) {
		float RotAlpha = PI/sampledOrder[s].size();
		std::vector<Cord> sides;
		int errV = dis(gen);

		for( int i=0 ; i<sampledOrder[s].size() ; i++ ) {
			Cord st = Cord(1,0);
			Cord dt = Cord(-1,0);
			int a1 = sampledOrder[s][i].first;
			int a2 = sampledOrder[s][i].second;

			float fromS = A.getDistance( a1, v );
			float fromD = A.getDistance( a2, v );
			float anchD = A.getDistance( a1, aA.getAnchor(a2) );

			if( a1 == errV ) {
				fromS += (fromD<= (fromS+anchD +1)) && (fromD>= std::abs(fromS-anchD+1));
				errV = dis(gen);
			}

			float cosalpha = -2;
			if( fromS==0 ) {
				cosalpha = 1;
			} else if( fromD==0 ) {
				cosalpha = -1;
			} else{
				cosalpha = (fromS*fromS+anchD*anchD-fromD*fromD)/(2*fromS*anchD);
			}

			float alpha = std::acos( cosalpha );

			float ratio = (DistEuclidien(st,dt)/anchD)*fromS;
			float x = st.first-std::cos(alpha)*ratio;
			float y = std::sin(alpha)*ratio;

			float nx = x*std::cos(i*RotAlpha) - y*std::sin(i*RotAlpha);
			float ny = x*std::sin(i*RotAlpha) + y*std::cos(i*RotAlpha);

			sides.push_back(Cord(nx,ny));
		}

		Cord pos = PolyCentroid(sides);

		return pos;
	}

	Cord trianglePosOnSamplesWErr( Vertex v ) {
		std::random_device rd;
		std::mt19937 gen(rd());
		std::uniform_int_distribution<> dis(0, orderedVAnchors.size()*2-1);

		std::vector<Cord> centroids;
		for( int i=0 ; i<sampledOrder.size() ; i++ ) {
			centroids.push_back( trianglePosWErr(v, i, dis, gen) );
		}
		float x=0, y=0;
		for( int i=0 ; i<centroids.size() ; i++ ) {
			auto c = centroids[i];
			x += c.first;
			y += c.second;
		}
		x = x/sampledOrder.size();
		y = y/sampledOrder.size();

		return Cord(x,y);
	}

	Cord PolyCentroid( std::vector<Cord>& l ) {
		polygon_type poly, poly_convex;
		point_type p;

		for( int i=0 ; i<l.size() ; i++ ) {
			auto c = l[i];
			boost::geometry::append( poly, point_type(c.first, c.second));
		}

		boost::geometry::centroid(poly, p);
		float x = boost::geometry::get<0>(p);
		float y = boost::geometry::get<1>(p);

		return Cord( x+2,y+1 );
	}

	Cord average( std::vector<Cord>& l ){
		float x=0, y=0;
		for( int i=0 ; i<l.size() ; i++ ) {
			auto c = l[i];
			x+=c.first;
			y+=c.second;
		}
		return Cord( x/l.size()+2, y/l.size() +1);
	}

	float DistEuclidien(Cord s, Cord d) {
		return (float)std::sqrt( (s.first-d.first)*(s.first-d.first) + (s.second-d.second)*(s.second-d.second) );
	}

};

#endif