#ifndef SIMILARITY_H
#define SIMILARITY_H

#include <vector>
#include <iostream>
#include <queue>
#include <algorithm>
#include <chrono>

#include "graph.hpp"
#include "anchors.hpp"

using namespace std::chrono;

template<typename Vertex, typename Edge, typename EType>
class PairSim{
private:
	using sGraph = Graph<Vertex, Edge, EType>;
	using sAnchors = Anchors<Vertex, Edge, EType>;
	using sPair = std::pair< Vertex, float >;
	using sPVector = std::vector< sPair >;
	using sMatrix = std::vector< sPVector >;
	using sMap = std::vector< std::pair<Vertex, float> >;

	sGraph A, B;
	sAnchors aA, aB;

#ifdef P_EXEC
	float tTau, tDelta, tCv, tCe, tTauV, tTauE, tAlpha, tSigma;
	high_resolution_clock::time_point t1,t2;

	float getInSec() {
		return ((float)duration_cast<microseconds>(t2-t1).count())/1000000;
	}
#endif

public:
	PairSim( ) { }

	PairSim( sGraph A, sGraph B, sAnchors aA, sAnchors aB ) {
		this->A = A;
		this->B = B;
		this->aA = aA;
		this->aB = aB;
#ifdef P_EXEC
	tTau=0; tDelta=0; tCv=0; tCe=0; tTauV=0; tTauE=0; tAlpha=0; tSigma=0;
#endif
	}

	void init( sGraph A, sGraph B, sAnchors aA, sAnchors aB ) {
		this->A = A;
		this->B = B;
		this->aA = aA;
		this->aB = aB;
#ifdef P_EXEC
	tTau=0; tDelta=0; tCv=0; tCe=0; tTauV=0; tTauE=0; tAlpha=0; tSigma=0;
#endif
	}

	float tau( Vertex v1,  Vertex v2) {
#ifdef P_EXEC
		t1 = high_resolution_clock::now();
#endif
		float sameT = ( A.getVType(v1) == B.getVType(v2) );

#ifdef P_EXEC
		t2 = high_resolution_clock::now();
		tTau += getInSec();
#endif
		return sameT;
	}

	float delta(Vertex v1, Vertex v2) {
#ifdef P_EXEC
		t1 = high_resolution_clock::now();
#endif
		float rdd=0;
		int d1 = A.getDeg( v1 );
		int d2 = B.getDeg( v2 );

		rdd = 2*std::abs( d1-d2 );
		rdd /= std::max( 2, d1+d2 );
		rdd += 1;
		rdd = 1./rdd;
#ifdef P_EXEC
		t2 = high_resolution_clock::now();
		tDelta += getInSec();
#endif
		return 1-rdd;
	}

	float tauV( Vertex v1, Vertex v2 ) {
#ifdef P_EXEC
		t1 = high_resolution_clock::now();
#endif

		float overlap = 0;
		int i=0, j=0;
		auto& w1 = A.getSVTypeList(v1);
		auto& w2 = B.getSVTypeList(v2);

		while( i<w1.size() && j<w2.size()) {
			if( w1[i]==w2[j] ) {
				overlap++;
				i++; j++;
			}
			else if( w1[i]<w2[j] ) {
				i++;
			}
			else {
				j++;
			}

		}
#ifdef P_EXEC
		t2 = high_resolution_clock::now();
		tTauV += getInSec();
#endif

		return overlap/( w1.size()+w2.size()-overlap );//nom/denom;
	}

	float tauE( Vertex v1, Vertex v2 ) {
#ifdef P_EXEC
		t1 = high_resolution_clock::now();
#endif
		float overlap = 0;
		int i=0, j=0;
		auto& w1 = A.getSETypeList(v1);
		auto& w2 = B.getSETypeList(v2);

		while( i<w1.size() && j<w2.size()) {
			if( w1[i]==w2[j] ) {
				overlap++;
				i++; j++;
			}
			else if( w1[i]<w2[j] ) {
				i++;
			}
			else {
				j++;
			}

		}
#ifdef P_EXEC
		t2 = high_resolution_clock::now();
		tTauE += getInSec();
#endif

		return overlap/( w1.size()+w2.size()-overlap );
	}

	float cV( Vertex v1, Vertex v2 )
	{
#ifdef P_EXEC
		t1 = high_resolution_clock::now();
#endif
		if( A.isAttributed() && B.isAttributed() ) {
			const std::string& max = A.getAttribute(v1);
			const std::string& min = B.getAttribute(v2);

			float overlap = 0;
			int i=0, j=0;
			while( i<min.length() && j<max.length()) {
				if( min[i]==max[j] ) {
					overlap++;
					i++; j++;
				}
				else if( min[i]<max[j] ) {
					i++;
				}
				else {
					j++;
				}
			}
#ifdef P_EXEC
		t2 = high_resolution_clock::now();
		tCv += getInSec();
#endif
			return overlap/( min.length()+max.length()-overlap );
		}
		else {
			return 0;
		}
	}

	float cE( Vertex v1, Vertex v2, float e ) {
#ifdef P_EXEC
		t1 = high_resolution_clock::now();
#endif
		int overlap = 0;
		auto& w1 = A.getSEWeightList(v1);
		auto& w2 = B.getSEWeightList(v2);

		float s1=0, s2=0;

		for( int k=0 ; k<w1.size() ; k++ ) {
			auto i = w1[k];
			for( int l=0 ; l<w2.size() ; l++ ) {
				auto j = w2[l];
				overlap += ( std::abs(i-j) < e );
			}
		}

#ifdef P_EXEC
		t2 = high_resolution_clock::now();
		tCe += getInSec();
#endif
		return overlap/( w1.size()*w2.size() );
	}

	float alpha( Vertex v1, Vertex v2, sMap& m, sMap& mR ) {
#ifdef P_EXEC
		t1 = high_resolution_clock::now();
#endif
		float nom = 0, denom =0 ;
		auto &alist1 = A.adj(v1);
		auto &alist2 = B.adj(v2);
		std::set<Vertex> aset1;

		for( int i=0 ; i<alist1.size() ; i++ ){
			auto&a = alist1[i];
			if( m[a.first].second > 1.0 ) {
				aset1.insert( m[a.first].first );
				denom++;
			}
		}

		for( int i=0 ; i<alist2.size() ; i++ ){
			auto&a = alist2[i];
			if( mR[a.first].second > 1.0 ) {
				denom++;
			}

			if( aset1.find(a.first)!=aset1.end() ) {
				++nom;
			}
		}
		denom -= nom;

#ifdef P_EXEC
		t2 = high_resolution_clock::now();
		tAlpha += getInSec();
#endif
		return ((denom==0) ? 0 : nom/denom);
	}

	float sigma( Vertex v1, Vertex v2, sMap& m, sMap& mR ) {
		return tau(v1, v2) * ( delta(v1, v2) + alpha(v1, v2, m, mR) + tauE(v1, v2) + tauV(v1, v2) + cE(v1, v2, 0) + 2*cV(v1, v2) )/6;
	}

	float sigmaBase( Vertex v1, Vertex v2 ) {
		return tau(v1, v2) * ( delta(v1, v2) + tauE(v1, v2) + tauV(v1, v2) + cE(v1, v2, 0) + cV(v1, v2) )/5;
	}

#ifdef P_EXEC
	void printExecTimes() {
		std::cout << "tau : " << tTau << std::endl;
		std::clog << "delta : " << tDelta << std::endl;
		std::clog << "alpha : " << tAlpha << std::endl;
		std::clog << "tauV : " << tTauV << std::endl;
		std::clog << "tauE : " << tTauE << std::endl;
		std::clog << "cV : " << tCv << std::endl;
		std::clog << "cE : " << tCe << std::endl;
	}
#endif
};


#endif